# See end of file for licensing information

package AppLib::HTML::Form;

# These are abstracted form primitive generators.
# They have an advantage for us over CGI.pm as they don't 
# integrate with queries and don't have default values.
# They just generate the HTML.

use strict;
use base 'AppLib::Error';
use Class::Autouse;

# Make sure the main HTML library is loaded
BEGIN { 
	# We will always need this, and right away too
	Class::Autouse->load( 'AppLib::HTML' ) 
}
	



 
#####################################################################
# Form componants

# Generate a textbox
sub textbox {
	my $class = shift;
	my $name = shift or return undef;
	my $value = shift;
	my $opt = shift || {};
	my %props = ( type => 'text', name => $name, value => $value );
	$value =~ s/$AppLib::HTML::property/$AppLib::HTML::escapeMap{$1}/g unless $opt->{dontescape};
	
	# Pass through options
	foreach ( qw{size maxlength class} ) {
		$props{$_} = $opt->{$_} if defined $opt->{$_};
	}
	return AppLib::HTML->tag( 'input', \%props );
}

# Generate a text area
sub textarea {
	my $class = shift;
	my $name = shift or return undef;
	my $content = shift;
	my $opt = shift || {};
	my %props = ( name => $name );
	$content =~ s/$AppLib::HTML::textarea/$AppLib::HTML::escapeMap{$1}/g unless $opt->{dontescape};

	# Pass through options
	foreach ( qw{cols rows class} ) {
		$props{$_} = $opt->{$_} if defined $opt->{$_};
	}
	
	# Handle special options
	$props{wrap} = uc $opt->{wrap} if defined $opt->{wrap};
	
	### FIXME - Add support for that weird IE html edit tag
	# $props{SOMETHING} = undef if exists $opt->{SOMETHING};
	
	return AppLib::HTML->tagPair( 'textarea', \%props, $content );
}

# Creates a drop box
sub dropbox {
	my $class = shift;
	my $name = shift or return undef;
	my $data = shift or return undef;
	my $value = shift;
	my $opt = shift || {};
	
	# Create the options array, and add all the data options
	my $havenull = 0;
	my @options = ();
	foreach ( @$data ) {
		$havenull = 1 if $_->[0] eq '';
		$_->[1] =~ s/$AppLib::HTML::text/$AppLib::HTML::escapeMap{$1}/g unless $opt->{dontescape};
		
		# Add the option to the array
		push @options, (defined $value and $value eq $_->[0])
			? "<option value=\"$_->[0]\" selected>$_->[1]</option>\n"
			: "<option value=\"$_->[0]\">$_->[1]</option>\n";
	}

	# Add the null entry if we can have it and do need it
	unless ( $opt->{notnull} || $havenull ) {
		$opt->{nullstring} = '' unless defined $opt->{nullstring};
		unshift @options, (defined $value and $value ne '')
			? "<option value=''>$opt->{nullstring}</option>\n"
			: "<option value='' selected>$opt->{nullstring}</option>\n";
	}
	
	# Now build the final tag and return
	my %props = ( name => $name );
	foreach ( qw{class onChange} ) {
		$props{$_} = $opt->{$_} if exists $opt->{$_};
	}
	return AppLib::HTML->tagPair( 'select', \%props, join( '', @options ) );
}

# Generate a cluster of radio buttons
# Create a set of radio buttons
sub radio {
	my $class = shift;
	my $name = shift or return undef;
	my $data = shift or return '';
	my $value = shift;
	my $layout = shift or return undef;
	my $opts = shift || {};

	# Map the value half of the data into radio buttons
	foreach ( @$data ) {
		if ( $_->[0] eq $value ) {
			$_->[0] =~ s/$AppLib::HTML::property/$AppLib::HTML::escapeMap{$1}/g unless $opts->{dontescape};
			$_->[0] = "<input type=\"radio\" name=\"$name\" value=\"$_->[0]\" checked>";
		} else {
			$_->[0] =~ s/$AppLib::HTML::property/$AppLib::HTML::escapeMap{$1}/g unless $opts->{dontescape};
			$_->[0] = "<input type=\"radio\" name=\"$name\" value=\"$_->[0]\">";
		}
	}
	
	# Now, hand them over to the appropriate layout method
	$layout = 'vertical' unless $layout;
	if ( $layout eq 'vertical' ) {
		return join "<br>", map { "$_->[0]$_->[1]" } @$data;
	
	} elsif ( $layout eq 'horizontal' ) {
		return join " ", map { $_->[0] . $_->[1] } @$data;
		
	} elsif ( $layout eq 'table' ) {
		# Generate the cell pairs
		my @cellpairs = ();
		my %cellprops = ref $opts->{td} ? %{$opts->{td}} : ();
		$cellprops{valign} = 'top' unless defined $cellprops{valign};
		$cellprops{nowrap} = undef if $opts->{nowrap};
		
		foreach ( @$data ) {
			push @cellpairs, "    "
				. AppLib::HTML->tagPair( 'td', \%cellprops, $_->[0] )
				. AppLib::HTML->tagPair( 'td', \%cellprops, $_->[1] );
		}
		
		# Format the cell pairs into a grid
		my @rows = ();
		my $cols = $opts->{cols} 
			? $opts->{cols}
			: ceil(scalar(@cellpairs) / $opts->{rows});
		while ( scalar @cellpairs ) {
			push @rows, join( '', map { "    $_\n" } splice(@cellpairs, 0, $opts->{cols}) );
		}
		
		# Generate the final table html
		my $table = join "\n", map { "  <tr>\n    $_\n  </tr>" } @rows;
		my %tableprops = ref $opts->{table} ? %{$opts->{table}} : ();
		$tableprops{border} = 0 unless defined $tableprops{border};
		$tableprops{cellspacing} = 0 unless defined $tableprops{cellspacing};
		$tableprops{cellpadding} = 1 unless defined $tableprops{cellpadding};
		return AppLib::HTML->tagPair( 'table', \%tableprops, $table );
	
	} elsif ( isa( $layout, 'CODE' ) ) {
		# Hand off to the external layout method
		return &{$layout}($data, $opts);
	
	} else { 
		return "Unknown radio button layout format '$layout'";
	}	
}

# Generate a checkbox
sub checkbox {
	my $class = shift;
	my $name = shift or return undef;
	my $value = shift;
	my $opt = shift || {};
	my %props = ( type => 'checkbox', name => $name, value => $value );
	$value =~ s/$AppLib::HTML::property/$AppLib::HTML::escapeMap{$1}/g unless $opt->{dontescape};
	
	# Pass through options
	foreach ( qw{class} ) {
		$props{$_} = $opt->{$_} if defined $opt->{$_};
	}
	if ( $opt->{checked} ) {
		$props{checked} = undef;
	} else {
		delete $props{checked};
	}
	return AppLib::HTML->tag( 'input', \%props );
}

# Create a hidden field
sub hidden {
	my $class = shift;
	my $name = shift or return undef;
	my $value = shift;
	
	# Escape the name and value
	$name = AppLib::HTML->escapeProperty( $name );
	$value = AppLib::HTML->escapeProperty( $value );
	
	# Generate the tag
	return AppLib::HTML->tag( 'input', { 
		type => 'hidden',
		name => $name,
		value => $value
		} );
}

1;

__END__

# Copyright (C) 2000-2002 Adam Kennedy ( software.applib@ali.as )
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.

