package CVSMonitor::Annotation;

# Object for parsing and statistical analysis of file annotations

use strict;
use UNIVERSAL 'isa';





#####################################################################
# Constructor 

# The constructor get's passed a single string containing the text
# for the annotation.
# If given, takes and stores an optional file name.
sub new {
	my $class = shift;
	my $content = shift or return undef;
	
	# Create the object
	my $self = {
		raw      => $content,
		lines    => [],
		};
	bless $self, $class;
	
	# Split the main string into lines
	my @lines = split /(?:\015\012|\015|\012)/, $content;
	
	foreach ( @lines ) {
		# Now find the line componants, and create a hash for them
		if ( /^([\d.]+)\s+\(([^\s\/\:]+)\s+([\w-]+)\)\:(.*)$/ ) {
			my $linehash = {
				revision => $1,
				author   => $2,
				date     => $3,
				content  => $4,
				};
			push @{ $self->{lines} }, $linehash;
		} else {
			# Bad line format
			return undef;
		}
	}
	
	return $self;
}





#####################################################################
# Access Methods

# Returns the raw annotation
sub raw { $_[0]->{raw} }

# Returns a reference to the lines
sub lines { [ @{ $_[0]->{lines} } ] }





#####################################################################
# Statistics

# Create and return a series of statistics based on revision, author
# and date.
sub stats {
	my $self = shift;
	
	# Create the stats struct
	my $stats = {
		revision => {},
		authors  => {},
		date     => {},
		};
	
	# Calculate the stats
	foreach ( @{ $self->{lines} } ) {
		$stats->{revision}->{ $_->{revision} }++;
		$stats->{authors}->{ $_->{author} }++;
		$stats->{date}->{ $_->{date} }++;
	}
	
	# Add the total number of lines, for convenience
	$stats->{lines} = scalar @{ $self->{lines} };
	
	return $stats;
}

1;

__END__

# Copyright (C) 2002 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
