package CVSMonitor::AuthorTotals;

# The CVSMonitor::AuthorTotals package implements a totals accumulator
# indexed by individual developers ( cvs accounts ).

use strict;
use UNIVERSAL qw{isa can};





#####################################################################
# Constructor and Static Methods

# Create a new developer index for a module
sub new {
	my $class = shift;
	my $Source = shift;
	my $balanced = shift;
	
	# Create the empty object
	my $self = {};
	bless $self, $class;

	# Add the module information if needed
	if ( $Source ) {
		$self->add( $Source, $balanced ) or return undef;
	}
	return $self;
}





#####################################################################
# Access Methods

# Returns the author hashes
sub getAuthors { [ values %{$_[0]} ] }

# Returns just the authors names
sub getAuthorNames { [ keys %{$_[0]} ] } 

# Get the authors sorted by a particular field
sub getAuthorsSorted {
	my $self = shift;
	my $value = lc shift;

	# Numeric, or author name
	if ( $value eq 'author' ) {
		return [ map { $self->{$_} } sort keys %$self ];
	}
	
	# Test the value
	my @authors = keys %$self;
	my $test = $self->getAuthorValue( $authors[0], $value );
	return undef unless defined $test;
	
	# Sort ( swartzian ) and return
	my @sorted = map { $self->{ $_->[0] } }
		sort { $a->[1] <=> $b->[1] }
		map { [ $_, $self->getAuthorValue( $_, $value ) ] }
		keys %$self;
	return \@sorted;
}

# Retrieves a particular statistic for the Author
sub getAuthorValue {
	my $self = shift;
	my $author = shift;
	my $value = lc shift;
	$author = $self->{$author} or return undef;
	
	# Is it an precalculated value
	if ( exists $author->{$value} ) {
		return $author->{$value};
	
	# Handle the other values we know about
	} elsif ( $value =~ /^average (\w+)$/ ) {
		return 0 if $author->{changes} == 0;
		return undef if $1 eq 'author';
		return 1 if $1 eq 'changes';
		
		if ( exists $author->{$1} ) {
			return int( ( $author->{$1} / $author->{changes} ) + 0.5 );
		} else {
			return undef;
		}
	} else {
		return undef;
	}
}

# Return a hash for all authors for a particular value
sub getValues {
	my $self = shift;
	my $value = shift;
	
	# Get the list of authors
	my @authors = keys %$self;
	return 0 unless scalar @authors;
	
	# Make sure the value is valid by testing against the first author;
	my $test = $self->getAuthorValue( $authors[0], $value );
	return undef unless defined $test;
	
	# Create and populate the author hash
	my $result = {};
	foreach ( @authors ) {
		$result->{$_} = $self->getAuthorValue( $_, $value );
	}
	
	return $result;
}

	
	


#####################################################################
# Main Methods

# Add Version information to the totals
sub add {
	my $self = shift;	
	my $Source = shift;
	my $balanced = shift;

	# Get the change sets
	return undef unless can( $Source, 'getChangeSets' );
	my $ChangeSets = $Source->getChangeSets or return undef;

	# Iterate through the change sets and add then
	foreach my $ChangeSet ( @$ChangeSets ) {		
		my $author = $ChangeSet->author;
		
		# Create the new author entry if it doesn't exist
		unless ( defined $self->{$author} ) {
			$self->{$author} = {
				author   => $author,
				changes  => 0,
				commits  => 0,
				add      => 0,
				remove   => 0,
				increase => 0,
				decrease => 0,
				karma   => 0,
				};
		}		
		
		# Add to the change count
		$self->{$author}->{commits}++;
		$self->{$author}->{changes} += $ChangeSet->changes;
		
		# Get the line metrics and add them
		my $lines = $balanced
			? $ChangeSet->getBalancedLines
			: $ChangeSet->getLines;
		foreach ( qw{add remove increase decrease karma} ) {
			$self->{$author}->{$_} += $lines->{$_};
		}
	}

	# Extend the Author basics with the extended values
	foreach ( values %$self ) {
		$_->{up} = $_->{add} + $_->{increase};
		$_->{down} = $_->{remove} + $_->{decrease};
		$_->{change} = $_->{up} - $_->{down};
		$_->{activity} = $_->{up} + $_->{down};
	}

	return 1;		
}

1;

__END__

# Copyright (C) 2002 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
