package CVSMonitor::DateDepth;

# This package implements a rough date with a specific date
# depth. ( Year, Month, Day )
#
# As input we only take ISO formatted date fragments.
# We only go as low as days.

use strict;
use UNIVERSAL 'isa';

# Resolution hash
use vars qw{@resolution %rhash};
BEGIN {
	push @resolution, { 
		isolate => '^(\d{4})',
		seperate => '^(\d{4})$', 
		depth => 'year', 
		};
	push @resolution, { 
		isolate => '^(\d{4}\/\d{2})', 
		seperate => '^(\d{4})\/(\d{2})$', 
		depth => 'month', 
		};
	push @resolution, { 
		isolate => '^(\d{4}\/\d{2}\/\d{2})', 
		seperate => '^(\d{4})\/(\d{2})\/(\d{2})$', 
		depth => 'day', 
		};
	%rhash = ();
	$rhash{$_->{depth}} = $_ foreach @resolution;
}




#####################################################################
# Constructor

sub new {
	my $class = shift;
	my $date = shift;

	my $self;
	if ( $date =~ /$resolution[0]->{seperate}/ ) {
		# Year
		$self = {			
			year  => $1,
			month => undef,
			day   => undef,
			depth => 0,
			};
	
	} elsif ( $date =~ /$resolution[1]->{seperate}/ ) {
		# Year/Month
		$self = {			
			year  => $1,
			month => $2,
			day   => undef,
			depth => 1,
			};
		
	} elsif ( $date =~ /$resolution[2]->{seperate}/ ) {
		# Year/Month/Day
		$self = {			
			year => $1,
			month => $2,
			day => $3,
			depth => 2,
			};
	
	} else {
		return $class->andError( "Unsupported date format" );
	}
	
	# Bless and return
	bless $self, $class;
	return $self;
}

# What level of depth is a particular date
sub getDepthOf {
	my $class = shift;
	my $string = shift;
	return 'year' if $string =~ /$resolution[0]->{seperate}/;
	return 'month' if $string =~ /$resolution[1]->{seperate}/;
	return 'day' if $string =~ /$resolution[2]->{seperate}/;
	return undef;
}

# Get the isolate and seperate regexs if you want them
sub isolate_regex {
	my $class = shift;
	my $depth = shift or return undef;
	my $resolution = $rhash{$depth} or return undef;
	return qr/$resolution->{isolate}/;
}
sub seperate_regex {
	my $class = shift;
	my $depth = shift or return undef;
	my $resolution = $rhash{$depth} or return undef;
	return qr/$resolution->{seperate}/;
}
	
sub year  { $_[0]->{year} }
sub month { $_[0]->{month} }
sub day   { $_[0]->{day} }
sub depth { $_[0]->{depth} }

# Get depth as a string
sub depthString {
	my $self = shift;
	if ( $self->{depth} == 0 ) {
		return 'year';
	} elsif ( $self->{depth} == 1 ) {
		return 'month';
	} elsif ( $self->{depth} == 2 ) {
		return 'day';
	} else {
		return undef;
	}
}

# Get the month including the year
sub fullMonth {
	my $self = shift;
	return undef unless $self->{depth} >= 1;
	return "$self->{year}/$self->{month}";
}
	





#####################################################################
# Increment and Decrement

# Increment the at our depth
sub increment {
	my $self = shift;
	if ( $self->{depth} == 0 ) {
		return $self->incrementYear;
	} elsif ( $self->{depth} == 1 ) {
		return $self->incrementMonth;
	} else {
		return $self->incrementDay;
	}
}

# Increment the year
sub incrementYear {
	my $self = shift;
	$self->{year}++;
	return 1;
}

# Increment the month
sub incrementMonth {
	my $self = shift;
	return undef unless $self->{depth} > 0;
	
	$self->{month}++;
	if ( $self->{month} > 12 ) {
		$self->{year}++;
		$self->{month} = 1;
	}
	
	return 1;
}

# Increment the day
sub incrementDay {
	my $self = shift;
	return undef unless $self->{depth} > 1;
	
	$self->{day}++;
	if ( $self->{day} > $self->_lastDay ) {
		$self->incrementMonth;
		$self->{day} = 1;
	}
	
	return 1;
}

# Decrement the date
sub decrement {
	my $self = shift;
	if ( $self->{depth} == 0 ) {
		return $self->decrementYear;
	} elsif ( $self->{depth} == 1 ) {
		return $self->decrementMonth;
	} else {
		return $self->decrementDay;		
	}
}
	
# Decrement the year
sub decrementYear {
	my $self = shift;
	$self->{year}--;
	return 1;
}

# Decrement the month
sub decrementMonth {
	my $self = shift;
	return undef unless $self->{depth};
	
	$self->{month}--;
	if ( $self->{month} == 0 ) {
		$self->{month} = 12;
		$self->{year}--;
	}
	
	return 1;
}

# Decrement the day
sub decrementDay {
	my $self = shift;
	return undef unless $self->{depth} > 1;
	
	$self->{day}--;
	if ( $self->{day} == 0 ) {
		$self->{month}--;
		$self->{day} = $self->_lastDay;
	}
	
	return 1;
}





#####################################################################
# Comparison

# Equality
sub equals {
	my $self = shift;
	my $rhs = shift;

	# Convert the RHS to an object if needed
	unless ( isa( $rhs, 'CVSMonitor::DateDepth' ) ) {
		$rhs = CVSMonitor::DateDepth->new( $rhs ) or die "Invalid Date comparison call";
	}

	# Check they are of the same depth
	unless ( $self->{depth} == $rhs->{depth} ) {
		die "Dates are of dissimilar depth";
	}
		
	# Do the equality check
	return ($self->toString eq $rhs->toString);
}
use overload '=' => \&equals;

# Less than
sub lessthan {
	my $self = shift;
	my $rhs = shift;
	my $reversed = shift;
	
	# Convert the RHS to an object if needed
	unless ( isa( $rhs, 'CVSMonitor::DateDepth' ) ) {
		$rhs = CVSMonitor::DateDepth->new( $rhs ) or die "Invalid Date comparison call";
	}
	
	# Check they are of the same depth
	unless ( $self->{depth} == $rhs->{depth} ) {
		die "Dates are of dissimilar depth";
	}
	
	# Fix reversal if required
	($self, $rhs) = ($rhs, $self) if $reversed;

	# Do the comparison.
	return ($self->toString lt $rhs->toString);
}
use overload '<' => \&lessthan;

# Greater than
sub greaterthan {
	my $self = shift;
	my $rhs = shift;
	my $reversed = shift;
	
	# Convert the RHS to an object if needed
	unless ( isa( $rhs, 'CVSMonitor::DateDepth' ) ) {
		$rhs = CVSMonitor::DateDepth->new( $rhs ) or die "Invalid Date comparison call";
	}
	
	# Check they are of the same depth
	unless ( $self->{depth} == $rhs->{depth} ) {
		die "Dates are of dissimilar depth";
	}
	
	# Fix reversal if required
	($self, $rhs) = ($rhs, $self) if $reversed;

	# Do the comparison
	return ($self->toString gt $rhs->toString);
}	
use overload '>' => \&greaterthan;
	
# Is a particular date "in" our date
sub isIn {
	my $self = shift;
	my $other = shift;
	my $string = $self->toString;
	return $other =~ /^$string/ ? 1 : 0;
}





#####################################################################
# Other usefull methods

sub toString {
	my $self = shift;
	return join '/', map { length($_) < 2 ? ('0'.$_) : $_ }
		grep { defined $_ }
		( $self->{year}, $self->{month}, $self->{day} );
}
use overload '""' => \&toString;

# Create a statistics totals generator
sub toTotalHash {
	my $self = shift;
	return {
		date => $self->toString,
		commits => 0,
		changes => 0,
		add => 0, 
		remove => 0,
		increase => 0,
		decrease => 0,
		karma => 0,
		};
}

# Find the highest and lowest dates in a module
sub findDateLimits {
	my $class = shift;
	my $Source = shift;
	my $depth = shift;

	# Check the depth
	my $regex = $class->isolate_regex( $depth ) or return undef;
	
	# Get the elements to find the range of
	my $Elements;
	if ( isa( $Source, 'CVSMonitor::MetaData::Activity' ) ) {
		$Elements = $Source;
	} elsif ( can( $Source, 'getChangeSets' ) ) {
		$Elements = $Source->getChangeSets;
	} elsif ( can( $Source, 'getVersions' ) ) {
		$Elements = $Source->getVersions;
	} else {
		return undef;
	}		

	# Populate the initial minimum and maximum values
	return undef unless scalar @$Elements;
	my $first = $Elements->[0];
	return undef unless $first->date =~ $regex;
	my ($min, $max) = ($1, $1);
	
	# Iterate over the versions
	foreach ( @$Elements ) {
		return undef unless $_->date =~ $regex;
		$min = $1 if $1 lt $min;
		$max = $1 if $1 gt $max;
	}
	
	return [ $min, $max ];
}

	



#####################################################################
# Utility methods

# Get the current day, month or year as an object
sub now {
	my $class = shift;
	my $depth = shift || 'month';
	
	# Get the date componants
	my @t = gmtime time;
	
	# Create an ISO formatted GMT date
	$t[4] += 1;
	$t[5] += 1900;
	$t[3] = '0' . $t[3] if length $t[3] < 2;
	$t[4] = '0' . $t[4] if length $t[4] < 2;

	if ( $depth eq 'day' ) {
		return $class->new( "$t[5]/$t[4]/$t[3]" );
	} elsif ( $depth eq 'month' ) {
		return $class->new( "$t[5]/$t[4]" );
	} elsif ( $depth eq 'year' ) {
		return $class->new( $t[5] );
	} else {
		return undef;
	}
}

# Find the last day of our current month
use vars qw{@lastDay};
BEGIN {
	@lastDay = qw{0 31 0 31 30 31 30 31 31 30 31 30 31};
}
sub _lastDay {
	my $self = shift;
	if ( $lastDay[ $self->{month} ] ) {
		return $lastDay[ $self->{month} ];
	}
	
	# Handle the rules for February
	return 28 unless $self->{month} % 4;
	return 29 unless $self->{month} % 100;
	return 28 unless $self->{month} % 400;
	return 29;
}

sub andError { $CVSMonitor::MetaData::errstr = $_[1]; undef }

1;

__END__

# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
