package CVSMonitor::MetaData::Abstract::Delta;

# This package provides an abstract parent class for both
# single Versions and ChangeSets.
# 
# It mainly provides for a common set of line related math
# that can be used for statistics purposes

# The main requirement for packages inheriting from this
# one is that the class defined a 'lines' property that
# has the following structure.

# {
#     add      => xxxx,
#     remove   => xxxx,
#     increase => xxxx,
#     decrease => xxxx,
# }

# This package provides the math to add the other, derived
# metrics.

use strict;





#####################################################################
# Constructor catcher

sub new {
	die "CVSMonitor::MetaData::Abstract::Delta is an abstract class. "
		. "You cannot instantiate it directly";
}





#####################################################################
# Main methods

# Fetch the common set of line metrics
sub getLines {
	my $self = shift;
	$self->_populateLineMetrics unless exists $self->{lines}->{change};
	
	# Return a mapping of the metrics
	my $lines = $self->{lines};
	return {
		add      => $lines->{add},
		remove   => $lines->{remove},
		increase => $lines->{increase},
		decrease => $lines->{decrease},
		up       => $lines->{up},
		down     => $lines->{down},
		activity => $lines->{activity},
		change   => $lines->{change},
		};
}

# Populate the line metric for a single Delta, 
# given add/remove/increase/decrease.
sub _populateLineMetrics {
	my $self = shift;
	my $lines = $self->{lines};
	
	# Define the simple metrics.
	# All metrics start as whole numbers.
	
	# "Up" is lines added plus lines increased.
	$lines->{up} = $lines->{add} + $lines->{increase};
	
	# "Down" is lines removed plus lines decreased
	$lines->{down} = $lines->{remove} + $lines->{decrease};
	
	# "Activity" is movement in any direction.
	# That is, "up" plus "down".
	$lines->{activity} = $lines->{up} + $lines->{down};
	
	# "Change" is the sum of all movement.
	# That is, "up" minus "down".
	$lines->{change} = $lines->{up} - $lines->{down};
	
	# Done
	return 1;
}





#####################################################################
# Methods to access the line values

sub add      { $_[0]->{lines}->{add} }
sub remove   { $_[0]->{lines}->{remove} }
sub increase { $_[0]->{lines}->{increase} }
sub decrease { $_[0]->{lines}->{decrease} }
sub activity { $_[0]->{lines}->{activity} }
sub change   { $_[0]->{lines}->{change} }

# Compatability methods
sub plus     { $_[0]->{lines}->{add} }
sub minus    { $_[0]->{lines}->{minus} }






#####################################################################
# Error handling

sub andError {
	$CVSMonitor::MetaData::errstr = $_[1];
	return undef;
}

1;


__END__

# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
