package CVSMonitor::MetaData::ChangeSet;

# A ChangeSet is a multi-file set of Versions made in the same module
# by the same author with gaps beween versions of less than 1 minute.

use strict;
use UNIVERSAL 'isa';
use base 'CVSMonitor::MetaData::Abstract::Delta';
use Sort::Versions ();

# Create a single delta from a list of Versions.
# Don't check stuff.
sub new {
	my $class = shift;

	# Get and check the versions
	my @Versions = @_;
	return undef unless scalar @Versions;
	foreach ( @Versions ) {
		return undef unless isa( $_, 'CVSMonitor::MetaData::Version' );
	}
	
	# Sort the version by filename and version
	@Versions = sort { 
		$a->File->filename cmp $b->File->filename
		||
		Sort::Versions::versioncmp( $a->revision, $b->revision )
		} @Versions;
	
	# Create and return the objects
	my $self = {
		id => undef,
		comment => $Versions[0]->comment,
		author => $Versions[0]->author,
		date => $Versions[0]->dateRaw,
		_VERSIONS => \@Versions,
		lines => {
			add      => 0,
			remove   => 0,
			increase => 0,
			decrease => 0,
			},
		};
	$self->{date_string} = $class->_get_date_string( $self->{date} );
	bless $self, $class;
	
	# Populate the line metrics
	$self->_populateLineMetrics or return undef;
	
	return $self;
}

# Isolate deltas from a ( potentially very large )
# set of Versions ( across multiple modules ).
# Take a reference to a list of Versions.
# Optionally takes a date to limit to.
# Return a refernece to a list of ChangeSets
sub findChangeSets {
	my $class = shift;
	my $V = isa( $_[0], 'ARRAY' ) ? shift : return undef;
	my $date = shift;

	# To save possibly huge amounts of isa calls, only check the first element
	return undef unless isa( $V->[0], 'CVSMonitor::MetaData::Version' );
	my @Versions = @$V;
	
	# Filter to the date if needed
	if ( $date ) {
		@Versions = grep { $_->inDate($date) } @Versions;
		return 0 unless scalar @Versions;
	}
	
	# Sort the Versions using a three level Schwartzian transform.
	# Sort by module, then author, then time.
	# Don't extract the value out of the arrayref yet, since
	# we will need the values later.
	my @Sorted = sort { $a->[1] cmp $b->[1] 
			or $a->[2] cmp $b->[2] 
			or $a->[3] cmp $b->[3]
			or $a->[4] <=> $b->[4] }
		map { [ $_, 
			$_->File->Module->name,
			$_->author,
			$_->comment,
			$_->dateRaw,
			] }
		@Versions;
		
	# Iterate through the sorted list, and group by
	# time and comment.
	my @ChangeSets = ();
	my @thisChangeSet = ();
	my $lastModule = '';
	my $lastAuthor = '';
	my $lastComment = '';	
	my $lastTime = 0;
	foreach my $element ( @Sorted ) {
		if ( $element->[1] ne $lastModule
			or $element->[2] ne $lastAuthor
			or $element->[3] ne $lastComment
			or ($element->[4] - $lastTime) > 120 
		) {
			# New ChangeSet, create and save the previous one.
			if ( scalar @thisChangeSet ) {
				my $ChangeSet = CVSMonitor::MetaData::ChangeSet->new( @thisChangeSet ) or return undef;
				push @ChangeSets, $ChangeSet;
			}
			
			# Update the variables
			@thisChangeSet = ();
			$lastModule = $element->[1];
			$lastAuthor = $element->[2];
			$lastComment = $element->[3];
		}
		
		# Add the current Version to the ChangeSet
		push @thisChangeSet, $element->[0];
		
		# Update the last time
		$lastTime = $element->[4];
	}
	
	# Create the last ChangeSet
	my $ChangeSet = CVSMonitor::MetaData::ChangeSet->new( @thisChangeSet ) or return undef;
	push @ChangeSets, $ChangeSet;

	# Sort the ChangeSets into a reliable order.
	@ChangeSets = sort { 
		$a->dateRaw <=> $b->dateRaw
		or
		$a->Module->name cmp $b->Module->name
		} @ChangeSets;
	
	# Tack on the ids
	my $id = 0;
	foreach ( @ChangeSets ) {
		$_->_set_id( ++$id );
	}
	
	# Return the ChangeSets
	return \@ChangeSets;	
}





#####################################################################
# Access Methods

sub id          { $_[0]->{id} }
sub author      { $_[0]->{author} }
sub dateRaw     { $_[0]->{date} }
sub comment     { $_[0]->{comment} }
sub commentSize { length $_[0]->{comment} }
sub date        { $_[0]->{date_string} }

# The number of changes ( versions in this case )
sub changes { scalar @{$_[0]->{_VERSIONS}} }

# Get the condensed string of the change
sub changeString {
	my $self = shift;
	my $balanced = shift;
	
	# Make sure we have the line metric
	$self->_populateLineMetrics unless exists $self->{lines}->{change};
	
	my ($plus, $minus);
	if ( $balanced ) {
		$plus = $self->{lines}->{balanced_up};
		$minus = $self->{lines}->{balanced_down};
	} else {
		$plus = $self->{lines}->{up};
		$minus = $self->{lines}->{down};
	}
	my $karma = $self->{lines}->{karma};
	my $merged = join ' ', ($karma ? "k$karma" : ()),
		 ($plus ? "+$plus" : ()),
		 ($minus ? "-$minus" : ());
	
	if ( $balanced and $self->{lines}->{balance_factor} ) {
		$plus = $self->{lines}->{up};
		$minus = $self->{lines}->{down};
		my $normal = join ' ', ($plus ? "+$plus" : ()), ($minus ? "-$minus" : ());
		return "$merged ($normal)";
	} else {
		return $merged;
	}
}

# Get to our various parents
sub Module     { $_[0]->{_VERSIONS}->[0]->Module }
sub Repository { $_[0]->{_VERSIONS}->[0]->Repository }
sub MetaData   { $_[0]->{_VERSIONS}->[0]->MetaData }

# Get the Versions in the ChangeSet
sub getVersions { @{ $_[0]->{_VERSIONS} } }

# Get the Files involved in the ChangeSet
sub getFiles {
	my $self = shift;

	# Prepare
	my @Files = ();
	my %hash = ();
	
	# Populate the hash, using the hash
	# to get the "distinct" Files.
	foreach my $File ( map { $_->File } @{$self->{_VERSIONS}} ) {
		next if $hash{$File};
		$hash{$File} = 1;
		push @Files, $File;
	}
	
	return \@Files;
}

# Is the ChangeSet within a certain date
sub inDate {
	my $self = shift;
	my $date = $self->date;
	return $date =~ /^$_[0]/ ? 1 : 0
}

sub _populateLineMetrics {
	my $self = shift;
	my $lines = $self->{lines};
	
	# First, total up the core values from the Versions
	foreach ( @{$self->{_VERSIONS}} ) {
		$lines->{add}      += $_->{lines}->{add};
		$lines->{remove}   += $_->{lines}->{remove};
		$lines->{increase} += $_->{lines}->{increase};
		$lines->{decrease} += $_->{lines}->{decrease};
	}
	
	# Do the remaining normal metrics
	$self->SUPER::_populateLineMetrics or return undef;
	
	# Determine the "balance factor".
	# This is the number of lines in "add" and "remove"
	# that cancel each other out.
	$lines->{balance_factor} = ($lines->{add} > $lines->{remove})
		? $lines->{remove} : $lines->{add};
	
	# Create the "balanced" versions of the above normal metrics.
	# It should not be necesary to explain them all.
	# The "change" value will be the same irrespective of balancing.
	$lines->{balanced_add} = $lines->{add} - $lines->{balance_factor};
	$lines->{balanced_remove} = $lines->{remove} - $lines->{balance_factor};
	$lines->{balanced_up} = $lines->{balanced_add} + $lines->{increase};
	$lines->{balanced_down} = $lines->{balanced_remove} + $lines->{decrease};
	$lines->{balanced_activity} = $lines->{balanced_up} + $lines->{balanced_down};

	# Calculate the 'karma' figures after all the normal stuff is known
	$lines->{karma} = $self->_karma();
	
	return 1;
}

# Get the Lines, but with the karma added
sub getLines {
	my $self = shift;
	my $lines = $self->SUPER::getLines();
	$lines->{karma} = $self->{lines}->{karma};
}

# Fetch the set of balanced metrics.
# "Balanced" in this case, means that the values
# for "add" and "remove" cancel each other out.
# The effect this has it that when files are renamed, or
# otherwise moved, in CVS ( which normally classifies them
# as seperate add/remove entries ), the "spike" will be 
# cancelled. The removal of the "false spikes" of activity
# should make for a much more accurate representation of 
# "real" addition and removal of files, and general 
# activity.
sub getBalancedLines {
	my $self = shift;
	$self->_populateLineMetrics unless exists $self->{lines}->{change};
	
	# Return a balanced mapping of the metrics
	my $lines = $self->{lines};
	return {
		add      => $lines->{balanced_add},
		remove   => $lines->{balanced_remove},
		increase => $lines->{increase},
		decrease => $lines->{decrease},
		up       => $lines->{balanced_up},
		down     => $lines->{balanced_down},
		activity => $lines->{balanced_activity},
		change   => $lines->{change},
		karma   => $lines->{karma},
		};
}

# Find the karma of a Change.
# Karma is a loose figure, based on three factors.
# 1. The number of files involved in the change
# 2. The amount of activity ( on balanced data )
# 3. The size of the comment for the commit
sub karma { 
	my $self = shift;
	$self->_populateLineMetrics unless exists $self->{lines}->{change};
	
	return $self->{lines}->{karma}
}

sub _karma {
	my $self = shift;
	
	# Get the number of files involved
	my $files_score = 0;
	foreach my $File ( @{ $self->getFiles } ) {
		if ( $File->binary ) {
			# Bigger score for binary files
			$files_score += 40;
		} else {
			$files_score += 15;
		}
	}
	
	# Get the amount of ( balanced activity )
	$self->_populateLineMetrics unless exists $self->{lines}->{change};
	my $activity = $self->{lines}->{balanced_activity};
	
	# To the normal activity we add half the net change, boosting the 
	# scores of large additions, and dropping the scores of large remocals
	$activity += $self->{lines}->{change};
	
	# Merge the points
	my $karma = $files_score + $activity + $self->commentSize;
	
	# To catch huge stupid mistakes, and giant imports, the maximum karma
	# score for a single commit is 10000. Any more than this, well, you 
	# should be committing in smaller chunks.
	return $karma > 10000 ? 10000 : $karma;
}	

sub _get_date_string {
	my $class = shift;
	my @t = gmtime shift;
	
	# Create an ISO formatted GMT date
	$t[4] += 1;
	$t[5] += 1900;
	$t[0] = '0' . $t[0] if length $t[0] < 2;
	$t[1] = '0' . $t[1] if length $t[1] < 2;
	$t[2] = '0' . $t[2] if length $t[2] < 2;
	$t[3] = '0' . $t[3] if length $t[3] < 2;
	$t[4] = '0' . $t[4] if length $t[4] < 2;
	return "$t[5]/$t[4]/$t[3] $t[2]:$t[1]:$t[0]";
}

sub _set_id {
	my $self = shift;
	$self->{id} = shift;
}

# Hopefully shut a warning up
sub DESTROY {}
		
1;

__END__

# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
