package CVSMonitor::MetaData::File;

# Implements information storage for a single file in
# a CVS repository. The focus is on creating all processor
# intensive work at the creation time, since this will likely
# be done in the background once, and the time of construction
# is not important.

use strict;
use UNIVERSAL 'isa';
use Sort::Versions ();

# In an effort to conserve memory, we are going to remove any version
# data that is not needed by the application in it's current version.
# As more features are added to the application, more keys can be kept.
# The following global contains the keys that we want to KEEP.
# Anything not in this list will be deleted at object creation time
use vars qw{%usable_keys};
BEGIN {
	%usable_keys = map { $_, 1 } qw{
		_PARENT
		_VERSIONS
		working_file
		keyword_substitution
		
		
		};
}




#####################################################################
# Constructor and Initialisation

# Create a new object from a fragment of CVS log test.
# An example of the text if provided at the end of this file
sub new {
	my $class = shift;
	my $lines = isa( $_[0], 'ARRAY' ) ? shift : return undef;
	my $parent = isa( $_[0], 'CVSMonitor::MetaData::Module' )
		? shift : return undef;
	chomp @$lines;
	
	# Create the object
	my $self = bless {
		_PARENT => $parent,
		_VERSIONS => undef,
		}, $class;
	
	# Populate the object properties
	$self->_initProperties($lines) or return undef;
	
	# Populate the initial, incomplete, versions
	$self->_initVersions($lines) or return undef;

	# Calculate the revisions
	$self->_completePrevious or return undef;
	
	# Clean up and finish the line metrics
	$self->_completeMetrics or return undef;

	# Clean out any unneeded values
	foreach ( keys %$self ) {
		delete $self->{$_} unless $usable_keys{$_};
	}

	return $self;
}

sub _initProperties {
	my $self = shift;
	my $lines = shift;
	
	# Parse them
	my ($key, $value);
	my $lastkey = 'none';
	foreach ( @$lines ) {
		next if /^\s*$/;
		last if /^description\:/;

		# Split into section based on ;
		foreach my $part ( split /\;\s+/, $_ ) {
			# If there is whitespace at the start
			# It's the child of the previous tag
			if ( $part =~ /^\s+([^:]+)\:\s*(.*)$/ ) {
				# Standardise the key format
				$key = lc $1; $value = $2;
				$key =~ s/\s/_/g;
				
				# Add the property to the last key
				unless ( ref $self->{$lastkey} ) {
					$self->{$lastkey} = { ':value:' => $self->{$lastkey} };
				}
				$self->{$lastkey}->{$key} = $value;				
			} elsif ( $part =~ /^([^:]+)\:\s*(.*)$/ ) {
				# Standardise the key format
				$key = lc $1; $value = $2;
				$key =~ s/\s/_/g;
								
				# Add the property to the object
				$self->{$key} = $value;
				$lastkey = $key;
			} else {
				return $self->andError( "Invalid log format" );
			}
		}		
	}
	return 1;
}

# Seperate out the version information and add them
sub _initVersions {
	my $self = shift;
	my $lines = shift;
	
	# Find the first single dash line
	my $state = 0;
	my @versionLines = ();
	my @thisVersion = ();
	foreach ( @$lines ) {
		if ( $state == 0 ) {
			# Before the first line
			next unless /^\-{28}$/;
			
			# Found the first version
			$state = 1;
			next;
		} else {
			if ( /^\-{28}$/ ) {
				# New version
				push @versionLines, [ @thisVersion ];
				@thisVersion = ();
			} else {
				# Add to the current version information
				push @thisVersion, $_;
			}
		}			
	}
	push @versionLines, [ @thisVersion ] if scalar @thisVersion;
	
	# Create the versions from the lines
	my @Versions = ();
	foreach my $lines ( @versionLines ) {
		my $Version = CVSMonitor::MetaData::Version->new( $lines, $self );
		return undef unless defined $Version;
		push @Versions, $Version;
	}

	# Always save the versions in ascending version order
	@Versions = sort { 
		Sort::Versions::versioncmp( $a->{revision}, $b->{revision} ) 
		} @Versions;
	$self->{_VERSIONS} = \@Versions;
	return 1;
}

# Go through the Versions and calculate the previous
# revision for each.
sub _completePrevious {
	my $self = shift;
	my @Versions = @{$self->{_VERSIONS}};

	# Sort the Versions by the revisions, and
	# build a index for use later.
	my %hash = ();
	$hash{$_->{revision}} = $_ foreach @Versions;

	# Go through the revisions in reverse order and work out 
	# each revision's previous version number.
	my @revisions = reverse map { $_->revision } @Versions;
	
	# Work out all the "greater than two parts" revisions using
	# the special revision math.
	foreach my $revision ( @revisions ) {
		# Skip the two digit revisions
		next if $revision =~ /^\d+\.\d+$/;
				
		# Step upwards through the tree until we find a revision that exists
		my $previous = $revision;
		while ( $previous = CVSMonitor::VersionNumber->previous( $previous ) ) {
			# Does this version exist?
			if ( exists $hash{$previous} ) {
				$hash{$revision}->{previous} = $previous;
				last; # Exit the while
			}
		}
		return undef unless defined $previous;
	}
	
	# Filter to get just the remaining 2 part revisions
	@revisions = grep { /^\d+\.\d+$/ } @revisions;
	for ( my $i = 0; $i < $#revisions; $i++ ) {
		$hash{ $revisions[$i] }->{previous} = $hash{ $revisions[$i + 1] }->{revision};
	}
	
	# Set the last previous value
	$hash{ $revisions[-1] }->{previous} = 0;
	return 1;
}

# Fill in some of the line metrics gaps
sub _completeMetrics {
	my $self = shift;
	my @Versions = @{$self->{_VERSIONS}};

	# Don't to line metrics for binary files
	unless ( $self->binary ) {
		# Go through the versions and set the "current" value
		my $lines;
		foreach my $Version ( @Versions ) {		
			
			# Get the previous revision
			my $Previous = $Version->previous;
			
			# Handle the first revision
			$lines = $Version->{lines};
			unless ( $Previous ) {
				# Current is the same as add
				$lines->{current} = $lines->{add};
				$Version->{action} = 'created';
				next;
			}
	
			# Do the math to set the current and remove
			if ( $Version->state eq 'Exp' ) {
				if ( $Previous->state eq 'dead' ) {
					$lines->{add} = $Previous->{lines}->{remove};
					$lines->{current} = $Previous->{lines}->{remove};
					$Version->{action} = 'created';
				} else {
					$lines->{current} = $Previous->{lines}->{current}
						+ $lines->{add}	+ $lines->{increase}
						- $lines->{decrease} - $lines->{remove};
				}
			} else {
				# Dead
				$lines->{remove} = $Previous->{lines}->{current};
				$lines->{current} = 0;
				$Version->{action} = 'removed';
			}
		}
	}

	# Populate the remaining metrics
	foreach ( @Versions ) {
		$_->_populateLineMetrics or return undef;
	}
	
	return 1;
}





#####################################################################
# Access methods

# Get the module and repository we are in
sub Module     { $_[0]->{_PARENT} }
sub Repository { $_[0]->{_PARENT}->Repository }
sub MetaData   { $_[0]->{_PARENT}->MetaData }

# Are we binary?
sub binary { $_[0]->{keyword_substitution} =~ /b/ ? 1 : 0 }

# The number of versions
sub versions { scalar @{ $_[0]->{_VERSIONS} } }

# Get the size of the file on disk ( in lines ) in the working area.
# Won't work for binary files.
sub _getLinesOnDisk {
	my $self = shift;
	return undef if $self->binary;
	return undef unless defined $self->{working_file};
	
	# Fully resolve the file name
	my $file = File::Spec->catfile( $self->Module->pathCheckoutDir, $self->{working_file} );
	return undef unless $file;
	
	# Does the file exist?
	unless ( -e $file ) {
		### TEMPORARY FIX
		# For the rare case that somebody DOESN'T start their versions at 1.1,
		# we add a workaround by saying that the first version has 0 lines.
		# This is technicall incorrect, and will lead to slightly corrupt
		# statistics, BUT means that the entire repository doesn't break.
		return 0;
	}
	
	# Open the file
	open ( LINESONDISK, $file ) or return undef;
	
	# Iterate through the file and count
	my $counter = 0;
	$counter++ while <LINESONDISK>;
	
	# Clean up and return the value
	close LINESONDISK;
	return $counter;
}
	
# Get an array ref of all the versions
sub getVersions { $_[0]->{_VERSIONS} }

# Get a particular version
sub getVersion {
	my $self = shift;
	my $revision = shift or return undef;
	
	# Search through the Versions for it
	foreach my $Version ( @{ $self->{_VERSIONS} } ) {
		if ( defined $Version->{revision} and $Version->{revision} eq $revision ) {
			return $Version;
		}
	}
	
	# Version not found
	return 0;
}

# Finds the first version.
# The first version will always be the one with the lowest
# revision number. Because these are sorted lowest up on 
# creation. The 'creation' version will be the first one
# in the array.
sub getFirstVersion {
	my $self = shift;
	return $self->{_VERSIONS}->[0] || 0;
}

# Get the most recent version.
# Because of the multiple branch thing, the most recent change
# might not be on the HEAD.
# So, find the one with the highest rawDate ( seconds past the epoch )
sub getRecentVersion {
	my $self = shift;
	
	my $Recent = $self->getFirstVersion;
	foreach my $Version ( @{ $self->getVersions } ) {
		if ( $Version->dateRaw > $Recent->dateRaw ) {
			$Recent = $Version;
		}
	}
	
	return $Recent;		
}

# Get the most recent HEAD version.
# This SHOULD be the version with only one . in it that is the
# highest
sub getRecentHeadVersion {
	my $self = shift;
	
	my $Recent = $self->getFirstVersion;
	foreach my $Version ( @{ $self->getVersions } ) {
		next unless $Version->revision =~ /^\d+\.\d+$/;
		if ( $Version->dateRaw > $Recent->dateRaw ) {
			$Recent = $Version;
		}
	}
	
	return $Recent;
}

# Get the local file name
sub filename { $_[0]->{working_file} }

# Pull a diff of the file from the repository
sub diff {
	my $self = shift;
	unless ( scalar @_ ) {
		return $self->andError( "No revision provided to diff against" );
	}
	
	# In the single argument form, get the diff for this version.
	# In the two argument version, get from one version to another version.
	my ($from, $to);
	if ( scalar @_ == 2 ) {
		$from = shift or return undef;
		$to = shift or return undef;

		# Make sure the two versions exist
		$self->getVersion( $from ) or return undef;
		$self->getVersion( $to ) or return undef;
	} elsif ( scalar @_ == 1 ) {
		$to = shift or return undef;
		$to = $self->getVersion( $to ) or return undef;
		$from = $to->previous
			or return $self->andError( "Cannot get previous revision for first version" );
		$to = $to->revision;
		$from = $from->revision;
	}	
	
	# Pull the diff from the repository
	my $command = "diff -u -r $from -r $to $self->{working_file}";
	return $self->Module->getCVSCommandResults( $command );
}

# Pull an annotated view of the file from the server
sub annotate {
	my $self = shift;

	# Get the annotated view of the most recent HEAD version
	my $Version = $self->getRecentHeadVersion or return undef;
	return $Version->annotate;
}

# Pull a contents of a version of the file from the repository
sub contents {
	my $self = shift;
	my $revision = $_[0] ? "-r $_[0]" : '';
	
	# Get the contents from CVS
	my $command = "checkout $revision -p $self->{working_file}";
	return $self->Module->getCVSCommandResults( $command );
}

# We treat this like the normal ->getLines method
sub getBalancedLines { $_[0]->getLines }
	




#####################################################################
# Utility methods

sub andError {
	$CVSMonitor::MetaData::errstr = $_[1];
	return undef;
}

1;

__END__

# Example of a file's CVS log entry

RCS file: /var/cvs/CPAN/applications/psp/batch.pl,v
Working file: applications/psp/batch.pl
head: 1.15
branch:
locks: strict
access list:
symbolic names:
keyword substitution: kv
total revisions: 15;    selected revisions: 15
description:
----------------------------
revision 1.15
date: 2002/03/08 17:03:32;  author: adam;  state: Exp;  lines: +1 -1
Minor extention change... keep their old extention and append
----------------------------
revision 1.14
date: 2002/02/12 16:47:45;  author: adam;  state: Exp;  lines: +1 -2
A bunch more stuff to make use strict; use warnings; use diagnostics happy
----------------------------
revision 1.13
date: 2002/02/12 15:59:48;  author: adam;  state: Exp;  lines: +3 -1
use warnings;
----------------------------
revision 1.12
date: 2002/02/12 15:17:03;  author: adam;  state: Exp;  lines: +3 -1
Turned on diagnostics and warnings
----------------------------
revision 1.11
date: 2002/02/10 13:51:05;  author: adam;  state: Exp;  lines: +1 -1
Don't give each file it's own directory ( yet ).
----------------------------
revision 1.10
date: 2002/02/10 13:30:00;  author: adam;  state: Exp;  lines: +1 -1
Save each file in a seperate directory
----------------------------
revision 1.9
date: 2002/02/10 13:11:04;  author: adam;  state: Exp;  lines: +7 -2
Minor fixes
----------------------------
revision 1.8
date: 2002/02/10 06:30:11;  author: adam;  state: Exp;  lines: +2 -2
Explicitly specify the root dir in saves
----------------------------
revision 1.7
date: 2002/02/10 06:24:35;  author: adam;  state: Exp;  lines: +3 -0
Remove trailing slashes
----------------------------
revision 1.6
date: 2002/02/10 06:20:05;  author: adam;  state: Exp;  lines: +3 -1
Find the modules properly
----------------------------
revision 1.5
date: 2002/02/10 06:11:05;  author: adam;  state: Exp;  lines: +5 -0
Added callback capability for _multicommand
----------------------------
revision 1.4
date: 2002/02/10 05:26:01;  author: adam;  state: Exp;  lines: +1 -1
Use htmlPage instead
----------------------------
revision 1.3
date: 2002/02/10 05:11:17;  author: adam;  state: Exp;  lines: +4 -1
Added minor fixes and checks
----------------------------
revision 1.2
date: 2002/02/10 05:09:58;  author: adam;  state: Exp;  lines: +2 -3
Moved the line number code into the HTML formatting module
----------------------------
revision 1.1
date: 2002/02/10 04:32:52;  author: adam;  state: Exp;
Added batch processing script






# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
