package CVSMonitor::MetaData::Version;

# Implements a specific version of a file.
# For the line metric, it inherits from the Delta abstract.

use strict;
use UNIVERSAL 'isa';
use base 'CVSMonitor::MetaData::Abstract::Delta';
use Class::Autouse qw{
	Time::ParseDate
	Sort::Versions
	};

# In an effort to conserve memory, we are going to remove any version
# data that is not needed by the application in it's current version.
# As more features are added to the application, more keys can be kept.
# The following global contains the keys that we want to KEEP.
# Anything not in this list will be deleted at object creation time
use vars qw{%usable_keys};
BEGIN {
	%usable_keys = map { $_, 1 } qw{
		_PARENT
		revision
		author
		date
		comment
		state
		action
		lines
		previous
		};
}





#####################################################################
# Constructor

# Objects are constructed from fragments of cvs log
# See example at end of file

sub new {
	my $class = shift;
	my $lines = isa( $_[0], 'ARRAY' ) ? shift : return undef;
	my $parent = isa( $_[0], 'CVSMonitor::MetaData::File' )
		? shift : return undef;
	my $filename = $parent->filename;
	
	# Create the empty object
	my $self = {
		_PARENT => $parent,
		};
	bless $self, $class;
	
	# Process the first, revision line
	my $revision = shift @$lines;
	unless ( $revision =~ /^revision ((?:\d+)(?:.\d+)+)\s*(.*)$/ ) {
		return $self->andError( "Invalid revision line in version lines" );
	}
	$self->{revision} = $1;
	
	# If we found anything after the revision number ( most likely a 'lock' )
	# on the revision, it SHOULD parse like a normal property, so prepend it to the
	# beginning of the next line, and let it get parsed normally.
	if ( defined $2 and length $2 ) {
		$lines->[0] = "$2 $lines->[0]";
	}
	
	# Second lines is a set of properties
	my @parts = split /\;\s*/, shift @$lines;
	foreach ( @parts ) {
		next if /^\s*$/;
		unless ( /^([^:]+)\:\s*(.*)$/ ) {
			return $self->andError( "Invalid log format at file '$filename', revision $self->{revision}" );
		}
		my $key = $1;
		my $value = $2;
		
		# Standardise the key format
		$key =~ s/\s/_/g;
		$key = lc $key;
		
		# Add the property to the object
		$self->{$key} = $value;
	}

	# Add a default author if none is provided
	unless ( defined $self->{author} and length $self->{author} ) {
		$self->{author} = '_unknown_';
	}

	# If the third line starts with the line 'branches:', it is in the format
	# branches: version; version;
	if ( $lines->[0] =~ s/^branches:\s+// ) {
		my @branches = split /\s+/, shift @$lines;
		foreach ( @branches ) {
			s/[^\d\.]//g;
			$self->{branches}->{$_} = 1;
		}
	}			
	
	# The rest of the lines are comment.
	# Strip leading and trailing whitespace off them.
	$self->{comment} = join "\n", @$lines;
	$self->{comment} =~ s/^\s+//s;
	$self->{comment} =~ s/\s+$//s;
	
	# Set the line information
	$self->_cleanLines() or return undef;

	# If there is a date, parse it into a Unix epoch seconds time
	if ( $self->{date} ) {
		Class::Autouse->load( 'Time::ParseDate' );
		$self->{date} = Time::ParseDate::parsedate( $self->{date} );
		unless ( $self->{date} ) {
			return $self->andError( "Error parsing date for revision '$self->{revision}' of file '$filename'" );
		}
	}

	# Trim off any object keys we don't need
	foreach ( keys %$self ) {
		delete $self->{$_} unless $usable_keys{$_};
	}

	return $self;
}	

sub _cleanLines {
	my $self = shift;
	
	# Handle binary files
	if ( $self->File->binary ) {
		$self->{lines} = {
			add      => 0,
			remove   => 0,
			increase => 0,
			decrease => 0,
			};
		return 1;
	}
	
	# Handle removed files
	if ( $self->{state} eq 'dead' ) {
		$self->{lines} = {
			add => 0,
			remove => undef, # We'll work this out later
			increase => 0,
			decrease => 0,
			};
			return 1;
	}
	
	# Handle the 1.1 version
	if ( $self->{revision} eq '1.1' ) {
		# Get the size of it on disk
		my $size = $self->File->_getLinesOnDisk;
		return undef unless defined $size;

		$self->{lines} = {
			add      => $size,
			remove   => 0,
			increase => 0,
			decrease => 0,
			};
		return 1;
	}
			
	# Is the file missing line information
	unless ( $self->{lines} ) {
		return $self->andError( "Line information missing from version $self->{revision}" );
	}
	
	# Read the lines text
	unless ( $self->{lines} =~ /^\+(\d+)\s+\-(\d+)$/ ) {
		return $self->andError( "Invalid line information in version $self->{revision}" );
	}
	
	$self->{lines} = {
		add      => 0,
		remove   => 0,
		increase => $1,
		decrease => $2,
		};
	return 1;
}

		



#####################################################################
# Access methods

sub revision    { $_[0]->{revision} }
sub author      { $_[0]->{author} }
sub state       { $_[0]->{state} }
sub comment     { $_[0]->{comment} }
# sub current     { $_[0]->{current} } # What does this value do?
sub commentSize { length $_[0]->{comment} }
sub changes     { 1 }
sub action      { $_[0]->{action} }
sub created     { (defined $_[0]->{action} and $_[0]->{action} eq 'created') ? 1 : 0 }
sub removed     { (defined $_[0]->{action} and $_[0]->{action} eq 'removed') ? 1 : 0 }

# Pass the date value through to the string generator
sub date        { CVSMonitor::MetaData->date( $_[0]->{date} ) }
sub dateRaw     { $_[0]->{date} }

# Get to our various parents
sub File       { $_[0]->{_PARENT} }
sub Module     { $_[0]->{_PARENT}->Module }
sub Repository { $_[0]->{_PARENT}->Repository }
sub MetaData   { $_[0]->{_PARENT}->MetaData }

# Get the Author of this version
sub Author {
	my $self = shift;
	return $self->Module->getAuthor( $self->{revision} );
}

# Get the ChangeSet we are part of
sub ChangeSet {
	my $self = shift;
	my $Module = $self->Module;
	
	foreach my $ChangeSet ( @{$Module->getChangeSets} ) {
		foreach my $Version ( $ChangeSet->getVersions ) {
			if ( $Version eq $self ) { # Tests same memory location, which they should be
				return $ChangeSet;
			}
		}
	}
	
	return undef;
}

# Version summary, include 'created' or 'removed'
sub summary {
	my $self = shift;
	return (defined $self->{action} and length $self->{action})
		? "$self->{revision} ($self->{action})"
		: $self->{revision};
}

# Get the condensed string of the change
sub changeString {
	my $self = shift;
	return 'binary' if $self->File->binary;
	my $plus = $self->{lines}->{up} ? "+$self->{lines}->{up}" : '';
	my $minus = $self->{lines}->{down} ? "-$self->{lines}->{down}" : '';
	return join ' ', grep { $_ } ($plus, $minus);
}

# Is the Version within a certain date
sub inDate {
	my $self = shift;
	my $date = $self->date;
	return $date =~ /^$_[0]/ ? 1 : 0
}

# Directly get the previous revision object
sub previous {
	my $self = shift;
	return $self->File->getVersion( $self->{previous} );
}

# Work out the 'next' revisions, down multiple branches
sub next {
	my $self = shift;
	my $Versions = $self->File->getVersions or return ();
	return () unless $Versions;
	
	# Get all the Versions that has us as the previous Versions
	my $revision = $self->revision;
	my @Next = grep { $_->{previous} eq $revision } @$Versions;
	return () unless @Next;
	
	# Sort by revision and return as a list
	Class::Autouse->load( 'Sort::Versions' );
	return sort { Sort::Versions::versioncmp( $a->{revision}, $b->{revision} ) } @Next;
}

# Pull the diff for this version against the previous one
sub diff {
	my $self = shift;
	
	# Try to get the existing copy
	my $cached = CVSMonitor::DataCache->get( $self, 'diff' );
	return $$cached if $cached;

	# Get the diff and cache it
	my $diff = $self->File->diff( $self->{revision} );
	return undef unless defined $diff;
	CVSMonitor::DataCache->set( $self, 'diff', $diff );
	return $diff;
}

# Pull an annotated view of the version from the server
sub annotate {
	my $self = shift;
	my $filename = $self->File->filename;
	
	# Get the annotation
	my $command = "annotate -r $self->{revision} $filename 2>/dev/null";
	return $self->Module->getCVSCommandResults( $command );
}

# Get an annotation object
sub Annotation {
	my $self = shift;

	# Get the annotation content
	my $content = $self->annotate or return undef;
	
	# Create a new Annotation object from it
	return CVSMonitor::Annotation->new( $content );
}

# Pull the contents for this file at this version
sub content {
	my $self = shift;
	return $self->File->contents( $self->{revision} );
}


# Hopefully shut a warning up
sub DESTROY {}

1;

__END__

# Example of a single revision

revision 1.12
date: 2002/02/12 15:17:03;  author: adam;  state: Exp;  lines: +3 -1
Turned on diagnostics and warnings





# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
