package CVSMonitor::Range;

# Provides an implementation of a range of variable depth
# dates. Basically, a functionality wrapper for an array
# of CVSMonitor::DateDepth objects.

use strict;
use UNIVERSAL qw{isa can};





#####################################################################
# Constructors

# Standard constructors.
# Takes as arguments the beginning and end dates.
sub new {
	my $class = shift;
	my $min = isa( $_[0], 'CVSMonitor::DateDepth' )
		? shift : CVSMonitor::DateDepth->new( shift );
	return undef unless defined $min;
	my $max = isa( $_[0], 'CVSMonitor::DateDepth' )
		? shift : CVSMonitor::DateDepth->new( shift );
	return undef unless defined $max;
	
	# Check the dates
	unless ( $min->depth == $max->depth ) {
		return $class->andError( "Dates are not of the same depth" );
	}
	if ( $max < $min ) {
		return $class->andError( "Maximum is greater than minimum" );
	}

	# Build the array
	my @dates = ( $class->getStatisticsStruct( $min) );
	while ( ! $min->equals( $max ) ) {
		$min->increment;
		push @dates, $class->getStatisticsStruct( $min );
	}

	# Create and return the Range
	return bless [ @dates ], $class;
}

# Create a range for a specific year
sub year {
	my $class = shift;
	my $year = shift or return undef;
	return undef unless $year =~ /^\d{4}$/;
	return $class->new( "$year/01", "$year/12" );
}

# Create a range for a specific month
sub month {
	my $class = shift;
	my $month = shift or return undef;
	$month = CVSMonitor::DateDepth->new( $month ) or return undef;
	my $from = $month->toString . '/01';
	my $to = $month->toString . '/' . $month->_lastDay;
	return $class->new( $from, $to );
}		

# Create a range to contain a source object.
# The optional second argument lists the depth.
sub newFor {
	my $class = shift;

	# Get the Versions
	my $Source = shift;
	return undef unless can( $Source, 'getChangeSets' );
	my $ChangeSets = $Source->getChangeSets or return undef;
	
	# Get the depth
	my $depth = shift || 'month';
	my $regex = CVSMonitor::DateDepth->isolate_regex( $depth ) or return undef;
	
	# Populate the initial minimum and maximum values
	return undef unless $ChangeSets->[0]->date =~ $regex;
	my ($min, $max) = ($1, $1);
	
	# Iterate over the versions
	foreach my $ChangeSets ( @$ChangeSets ) {
		return undef unless $ChangeSets->date =~ $regex;
		$min = $1 if $1 lt $min;
		$max = $1 if $1 gt $max;
	}
	
	# Now we have the first and last dates, create the new Range for it
	return $class->new( $min, $max );
}

	



#####################################################################
# Access Methods

sub first { $_[0]->[0] }
sub last { $_[0]->[-1] }
sub depth { CVSMonitor::DateDepth->getDepthOf( $_[0]->[0]->{date} ) }

# Find the maximum value of a particular numeric field
sub max {
	my $self = shift;
	my $field = shift;
	return undef unless exists $self->first->{$field};
	
	# Set the default value
	my $max = $self->first->{$field};
	
	# Iterate over the elements
	foreach my $Date ( @$self ) {
		$max = $Date->{$field} if $Date->{$field} > $max;
	}
	
	return $max;
}

# Find the minimum value of a particular numeric field
sub min {
	my $self = shift;
	my $field = shift;
	return undef unless exists $self->first->{$field};
	
	# Set the default value
	my $min = $self->first->{$field};
	
	# Iterate over the elements
	foreach my $Date ( @$self ) {
		$min = $Date->{$field} if $Date->{$field} < $min;
	}
	
	return $min;
}

# "Slice" across a field.
# For example, given the field name 'blah', will return
# a reference to an array of all the 'blah' values.
sub slice {
	my $self = shift;
	my $field = shift;
	return undef unless exists $self->first->{$field};
	
	# Create and return the slice
	my @slice = map { $_->{$field} } @$self;
	return \@slice;
}





#####################################################################
# Functional Methods

# Add ( merge ), this Range with one or more others.
# Returns a new, merged, Range.
sub merge {
	my $self = shift;
	return $self unless scalar @_;	
	my $class = ref $self;
	my @Ranges = ($self, @_);
	
	# Find the maximum and minimum across the ranges
	my $min = $Ranges[0]->first->{date};
	my $max = $Ranges[0]->last->{date};
	foreach my $Range ( @Ranges ) {
		$min = $Range->first->{date} if $Range->first->{date} lt $min;
		$max = $Range->last->{date} if $Range->last->{date} gt $max;
	}		
	
	# Create the new range and index it
	my $Merged = $class->new( $min, $max ) or return undef;
	my %hash = ();
	$hash{$_->{date}} = $_ foreach @$Merged;
			
	# Now add the totals from each to the merged range
	foreach my $Range ( @Ranges ) {
		foreach my $stats ( @$Range ) {
			my $key = $stats->{date};
			next unless exists $hash{$key};
	
			# Add to the hash
			foreach ( keys %{ $hash{$key} } ) {
				next unless defined $hash{$key}->{$_};
				next if $_ eq 'date';
				next if ref $hash{$key}->{$_};
				unless ( defined $stats->{$_} ) {
					$hash{$key}->{$_} = undef;
					next;
				}
				$hash{$key}->{$_} += $stats->{$_};
			}
		}
	}

	return $Merged;
}
	
# Add Version information to the Range totals
sub add {
	my $self = shift;
	my $class = ref $self;
	die "Class $class does not implement the 'add' method";
}	
	
# Merge the statistics into totals for the entire range
sub totals {
	my $self = shift;
	my $class = ref $self;
	die "Class $class does not implement the 'totals' method";
}

# Extend the Range boundary to a particular date, in either direction
# as needed.
sub extend {
	my $self = shift;
	my $class = ref $self;

	# Check the new date
	my $extend_to = isa( $_[0], 'CVSMonitor::DateDepth' )
		? shift : CVSMonitor::DateDepth->new( shift );
	return undef unless defined $extend_to;
	
	# Is the date of the same depth
	unless ( $self->depth eq $extend_to->depthString ) {
		return $self->andError( "Date to extend to is of a different depth than the Range" );
	}
	
	# Is the Date before the start
	my $Extend = undef;
	if ( $extend_to < $self->first->{date} ) {
		# Create the new range
		$Extend = $class->new( $extend_to, $self->first->{date} ) or return undef;
	} elsif ( $extend_to > $self->last->{date} ) {
		$Extend = $class->new( $self->last->{date}, $extend_to ) or return undef;
	} else {
		# Don't need to extend
		return $self;
	}
	
	# Merge in the new bit
	return $self->merge( $Extend ) || undef;
}

# Extend as need to encompass today's date.
sub extendToNow {
	my $self = shift;
	
	# Get todays date of the appropriate depth
	my $depth = $self->depth;
	my $now = CVSMonitor::DateDepth->now( $depth ) or return undef;
	
	# Extend to it
	return $self->extend( $now );
}

# For a given date, generate it's basic stats struct
sub getStatisticsStruct {
	my $either = shift;
	my $date = shift;
	return undef unless isa( $date, 'CVSMonitor::DateDepth' );
	
	return {
		date => $date->toString,
		};	
}

1;
 

__END__

# Copyright (c) 2002 Adam Kennedy. All rights reserved.
# This program is free software; you can redistribute
# it and/or modify it under the same terms as Perl itself.
#
# The full text of the license can be found in the
# LICENSE.PERL file included with this module.
